#pragma OPENCL EXTENSION cl_khr_byte_addressable_store : enable

#define u32 uint
#define u8 uchar
#define size_t uchar
#define rol(value, bits) (((value) << (bits)) | ((value) >> (32 - (bits))))

/* blk0() and blk() perform the initial expand. */
/* I got the idea of expanding during the round function from SSLeay */
#define blk0(i) (block->l[i] = (rol(block->l[i], 24) & 0xFF00FF00) | (rol(block->l[i], 8) & 0x00FF00FF))
#define blk(i) (block->l[i & 15] = rol(block->l[(i + 13) & 15] ^ block->l[(i + 8) & 15] ^ block->l[(i + 2) & 15] ^ block->l[i & 15], 1))

/* (R0+R1), R2, R3, R4 are the different operations used in SHA1 */
#define R0(v,w,x,y,z,i) z += ((w & (x ^ y)) ^ y) + blk0(i) + 0x5A827999 + rol(v, 5); w = rol(w, 30);
#define R1(v,w,x,y,z,i) z += ((w & (x ^ y)) ^ y) + blk(i) + 0x5A827999 + rol(v, 5); w = rol(w, 30);
#define R2(v,w,x,y,z,i) z += (w ^ x ^ y) + blk(i) + 0x6ED9EBA1 + rol(v, 5); w = rol(w, 30);
#define R3(v,w,x,y,z,i) z += (((w | x) & y) | (w & x)) + blk(i) + 0x8F1BBCDC + rol(v, 5); w = rol(w, 30);
#define R4(v,w,x,y,z,i) z += (w ^ x ^ y) + blk(i) + 0xCA62C1D6 + rol(v, 5); w=rol(w, 30);


#define get_global_id(a) 0

constant size_t SHA1_MAC_LEN = 20;

/* compare the private memory block to the constant memory block
 * dest: the private memory block to compare
 * src: the constant memory block to compare
 * amount: the amount of bytes to compare
 * constoffset: the offset on the const block.
 * returns: 0 if it is not equal, 1 otherwise
 **/
inline char os_memcmp(private const void* dest, constant const void* src, const int amount) {
    char j;
    const int* destination = (const int*) dest;
    constant int* source = (constant int*) src;
    if(amount == 0) return 0;
    for(j = (amount >> 2) - 1; j >= 0; j--) if(destination[j] != source[j]) return 0;
    return 1;
}

inline char os_memcpy(private void* dest, private const void* src, const uchar amount) {
    uchar j = 0, intamount = amount >> 2;
    int* destination = (int*) dest;
    const int* source = (const int*) src;
    char* destination2 = (char*) dest;
    const char* source2 = (const char*) src;
    if(amount == 0) return 0;
    while(j < intamount) {
        destination[j] = source[j];
        j++;
    }
    j = j << 2;
    while(j < amount) {
        destination2[j] = source2[j];
        j++;
    }
    return 1;
}

inline void os_memcpyc(private void* dest, constant const void* src, const uchar amount) {
    uchar j = 0, intamount = amount >> 2;
    int* destination = (int*) dest;
    constant int* source = (constant int*) src;
    char* destination2 = (char*) dest;
    constant char* source2 = (constant char*) src;
    if(amount == 0) return;
    while(j < intamount) {
        destination[j] = source[j];
        j++;
    }
    j = j << 2;
    while(j < amount) {
        destination2[j] = source2[j];
        j++;
    }
}

inline void os_memcpywithIDs(private void* dest, constant const void* src, const uchar amount) {
    constant char* reflector = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcedfghijklmnopqrstuvwxyz?@";
    uchar j;
    uint* dst = (uint*) dest;
    constant uint* source = (constant uint*) src;
    for(j = (amount >> 2) - 1; j > 0; j--) dst[j] = source[j];
    dst[0] = (reflector[(get_global_id(0) & 63)] | (reflector[(get_global_id(1) & 63)] << 8) | (reflector[(get_global_id(2) & 63)] << 16) | (reflector[((get_global_id(0) >> 6) | ((get_global_id(1) >> 4) & 0xC) | ((get_global_id(2) >> 2) & 0x30))] << 24));
}
	
/* Hash a single 512-bit block. This is the core of the algorithm. */

void SHA1Transform(private unsigned char buffer[64], private uint* ctx)
{
	unsigned int a, b, c, d, e;
	typedef union {
		unsigned char c[64];
		unsigned int l[16];
	} CHAR64LONG16;
	CHAR64LONG16* block;
	block = (CHAR64LONG16 *) buffer;
	/* Copy context->state[] to working vars */
	a = ctx[0];
	b = ctx[1];
	c = ctx[2];
	d = ctx[3];
	e = ctx[4];
	/* 4 rounds of 20 operations each. Loop unrolled. */
	R0(a,b,c,d,e, 0); R0(e,a,b,c,d, 1); R0(d,e,a,b,c, 2); R0(c,d,e,a,b, 3);
	R0(b,c,d,e,a, 4); R0(a,b,c,d,e, 5); R0(e,a,b,c,d, 6); R0(d,e,a,b,c, 7);
	R0(c,d,e,a,b, 8); R0(b,c,d,e,a, 9); R0(a,b,c,d,e,10); R0(e,a,b,c,d,11);
	R0(d,e,a,b,c,12); R0(c,d,e,a,b,13); R0(b,c,d,e,a,14); R0(a,b,c,d,e,15);
	R1(e,a,b,c,d,16); R1(d,e,a,b,c,17); R1(c,d,e,a,b,18); R1(b,c,d,e,a,19);
	R2(a,b,c,d,e,20); R2(e,a,b,c,d,21); R2(d,e,a,b,c,22); R2(c,d,e,a,b,23);
	R2(b,c,d,e,a,24); R2(a,b,c,d,e,25); R2(e,a,b,c,d,26); R2(d,e,a,b,c,27);
	R2(c,d,e,a,b,28); R2(b,c,d,e,a,29); R2(a,b,c,d,e,30); R2(e,a,b,c,d,31);
	R2(d,e,a,b,c,32); R2(c,d,e,a,b,33); R2(b,c,d,e,a,34); R2(a,b,c,d,e,35);
	R2(e,a,b,c,d,36); R2(d,e,a,b,c,37); R2(c,d,e,a,b,38); R2(b,c,d,e,a,39);
	R3(a,b,c,d,e,40); R3(e,a,b,c,d,41); R3(d,e,a,b,c,42); R3(c,d,e,a,b,43);
	R3(b,c,d,e,a,44); R3(a,b,c,d,e,45); R3(e,a,b,c,d,46); R3(d,e,a,b,c,47);
	R3(c,d,e,a,b,48); R3(b,c,d,e,a,49); R3(a,b,c,d,e,50); R3(e,a,b,c,d,51);
	R3(d,e,a,b,c,52); R3(c,d,e,a,b,53); R3(b,c,d,e,a,54); R3(a,b,c,d,e,55);
	R3(e,a,b,c,d,56); R3(d,e,a,b,c,57); R3(c,d,e,a,b,58); R3(b,c,d,e,a,59);
	R4(a,b,c,d,e,60); R4(e,a,b,c,d,61); R4(d,e,a,b,c,62); R4(c,d,e,a,b,63);
	R4(b,c,d,e,a,64); R4(a,b,c,d,e,65); R4(e,a,b,c,d,66); R4(d,e,a,b,c,67);
	R4(c,d,e,a,b,68); R4(b,c,d,e,a,69); R4(a,b,c,d,e,70); R4(e,a,b,c,d,71);
	R4(d,e,a,b,c,72); R4(c,d,e,a,b,73); R4(b,c,d,e,a,74); R4(a,b,c,d,e,75);
	R4(e,a,b,c,d,76); R4(d,e,a,b,c,77); R4(c,d,e,a,b,78); R4(b,c,d,e,a,79);
	/* Add the working vars back into context.state[] */
	ctx[0] += a;
	ctx[1] += b;
	ctx[2] += c;
	ctx[3] += d;
	ctx[4] += e;
}

/* SHA1Init - Initialize new context */

inline void SHA1Init(private uint* ctx)
{
	/* SHA1 initialization constants */
	ctx[0] = 0x67452301;
    ctx[1] = 0xEFCDAB89;
    ctx[2] = 0x98BADCFE;
    ctx[3] = 0x10325476;
    ctx[4] = 0xC3D2E1F0;
    ctx[5] = 0;
    ctx[6] = 0;
}

void SHA1Update(private unsigned char* data, private const uchar len, private uint* ctx, private uchar* ctxbuffer)
{
	u32 i, j;
	j = (ctx[5] >> 3) & 63;
	if ((ctx[5] += len << 3) < (len << 3))
		ctx[6]++;
	ctx[6] += (len >> 29);
	if ((j + len) > 63) {
		os_memcpy(&ctxbuffer[j], data, (i = 64-j));
		SHA1Transform((private uchar*)ctxbuffer, ctx);
		for ( ; i + 63 < len; i += 64) {
			SHA1Transform(&data[i], ctx);
		}
		j = 0;
	}
	else i = 0;
	os_memcpy(&ctxbuffer[j], &data[i], len - i);
}

void SHA1Updatec(constant unsigned char* data, const uchar len, uint* private ctx, uchar* private ctxbuffer)
{
	u32 i,j;
	j = (ctx[5] >> 3) & 63;
	if ((ctx[5] += len << 3) < (len << 3))
		ctx[6]++;
	ctx[6] += (len >> 29);
   	if ((j + len) > 63) {
		os_memcpyc((private uchar*) &ctxbuffer[j], data, (private int) (i = 64-j));
		SHA1Transform((private uchar*)ctxbuffer, (private uint*) ctx);
		j = 0;
	}
	else i = 0;
	os_memcpyc((private uchar*) &ctxbuffer[j], &data[i], len - i);
}


/* Add padding and return the message digest. */

void SHA1Final(private unsigned char digest[20], private uint* ctx, private uchar* ctxbuffer)
{
	u32 i;
	unsigned char finalcount[8];
    
	        for (i = 0; i < 8; i++) {
                finalcount[i] = (unsigned char)
                        ((ctx[(i >= 4 ? 5 : 6)] >>
                          ((3-(i & 3)) * 8) ) & 255);  /* Endian independent */
        }
        SHA1Updatec((constant unsigned char *) "\200", 1, ctx, ctxbuffer);
        while ((ctx[5] & 504) != 448) {
                SHA1Updatec((constant unsigned char *) "\0", 1, ctx, ctxbuffer);
        }
        SHA1Update(finalcount, 8, ctx, ctxbuffer);  /* Should cause a SHA1Transform() */
        for (i = 0; i < 20; i++) {
                digest[i] = (unsigned char)
                        ((ctx[i >> 2] >> ((3 - (i & 3)) * 8)) & 255);
        }

}

int sha1_vector(const size_t num_elem2, private u8 *addr2[], private size_t* len2, const size_t num_elem, constant u8 *addr[], const size_t *len, private u8 *mac)
{
    uint i;
    private uint ctx[7];
    private uchar ctxbuffer[64];
	SHA1Init(ctx);
    SHA1Update((private u8*)addr2[0], len2[0], ctx, ctxbuffer);
        for (i = 0; i < num_elem; i++)
		SHA1Updatec((constant u8*)addr[i], (size_t)len[i], ctx, ctxbuffer);
	for (i = 1; i < num_elem2; i++)
		SHA1Update((private u8*)addr2[i], len2[i], ctx, ctxbuffer);
    SHA1Final(mac, ctx, ctxbuffer);
    return 0;
}

/**
 * hmac_sha1_vector - HMAC-SHA1 over data vector (RFC 2104)
 * @key: Key for HMAC operations
 * @key_len: Length of the key in bytes
 * @num_elem: Number of elements in the data vector
 * @addr: Pointers to the data areas
 * @len: Lengths of the data blocks
 * @mac: Buffer for the hash (20 bytes)
 * Returns: 0 on success, -1 on failure
 */
int hmac_sha1_vector(constant const char *key, size_t key_len, size_t num_elem, size_t num_elem2, constant u8 **addr, private u8 *addr2, const size_t *len, private const size_t *len2, private u8 *mac)
{
	private unsigned char k_pad[64]; /* padding - key XORd with ipad/opad */
	unsigned char tk[20];
    uchar* tmp = mac;
	constant u8 *_addr[2];
    private u8 *_addr2[2];
	size_t i;
    private size_t _len2[2];
	int ret;

	/* the HMAC_SHA1 transform looks like:
	 *
	 * SHA1(K XOR opad, SHA1(K XOR ipad, text))
	 *
	 * where K is an n byte key
	 * ipad is the byte 0x36 repeated 64 times
	 * opad is the byte 0x5c repeated 64 times
	 * and text is the data being protected */

	/* start out by storing key in ipad */
	os_memcpywithIDs(k_pad, key, key_len);
    for(i = key_len; i < 64; i++) k_pad[i] = 0;
    
	/* XOR key with ipad values */
	for (i = 0; i < 64; i++)
		k_pad[i] ^= 0x36;

	/* perform inner SHA1 */
	_addr2[0] = k_pad;
	_len2[0] = 64;
	for (i = 0; i < num_elem; i++) {
		_addr[i] = addr[i];
	}
    for (i = 0; i < num_elem2; i++) {
		_addr2[i + 1] = addr2;
		_len2[i + 1] = len2[i];
	}
	if (sha1_vector(1 + num_elem2, _addr2, _len2, num_elem, _addr, len, mac))
		return -1;
	os_memcpywithIDs(k_pad, key, key_len);
	for(i = key_len; i < 64; i++) k_pad[i] = 0;
    /* XOR key with opad values */
	for (i = 0; i < 64; i++)
		k_pad[i] ^= 0x5c;

	/* perform outer SHA1 */
	_addr2[0] = k_pad;
	_len2[0] = 64;
	_addr2[1] = mac;
	_len2[1] = SHA1_MAC_LEN;
	ret = sha1_vector(2, _addr2, _len2, 0, _addr, len, mac);
	return ret;
}


/**
 * hmac_sha1 - HMAC-SHA1 over data buffer (RFC 2104)
 * @key: Key for HMAC operations
 * @key_len: Length of the key in bytes
 * @data: Pointers to the data area
 * @data_len: Length of the data area
 * @mac: Buffer for the hash (20 bytes)
 * Returns: 0 on success, -1 of failure
 */
int hmac_sha1(constant const char *key, const size_t key_len, private u8 *data, size_t data_len, private u8 *mac)
{
	return hmac_sha1_vector(key, key_len, 0, 1, (constant uchar**) &key, data, &key_len, &data_len, mac); //2 last '&key' uses should be subsituted by NULL, but this is not allowed, so therefore this dirty hack
}

int pbkdf2_sha1_f(constant const char *passphrase, size_t passphrase_len, constant uchar *ssid,
			 size_t ssid_len, const int iterations, unsigned int count,
			 private u8 *digest)
{
	private unsigned char tmp[SHA1_MAC_LEN], tmp2[SHA1_MAC_LEN];
	int i, j;
	private uchar count_buf[4];
	const size_t len = 4;
    u8 addr[2];
/*
	addr[0] = (uchar*) ssid;
	len[0] = ssid_len;
	addr[1] = (uchar*) count_buf;
	len[1] = 4;
*/
	/* F(P, S, c, i) = U1 xor U2 xor ... Uc
	 * U1 = PRF(P, S || i)
	 * U2 = PRF(P, U1)
	 * Uc = PRF(P, Uc-1)
	 */

	count_buf[0] = (count >> 24) & 0xff;
	count_buf[1] = (count >> 16) & 0xff;
	count_buf[2] = (count >> 8) & 0xff;
	count_buf[3] = count & 0xff;
	if (hmac_sha1_vector(passphrase, passphrase_len, 1, 1, &ssid, count_buf, &ssid_len, &len, tmp))
		return -1;

    os_memcpy(digest, tmp, SHA1_MAC_LEN);

	for (i = 1; i < iterations; i++) {
		if (hmac_sha1(passphrase, passphrase_len, tmp,
			      SHA1_MAC_LEN, tmp2))
			return -1;
		os_memcpy(tmp, tmp2, SHA1_MAC_LEN);
		for (j = 0; j < SHA1_MAC_LEN; j++)
			digest[j] ^= tmp2[j];
	}
	return 0;
}

char pbkdf2_sha1(constant char *passphrase, size_t length, constant uchar *ssid, size_t ssid_len, constant char* pos)
{
	unsigned char count = 0, result = 0;
	uchar left = 32, plen;
	private unsigned char digest[80];

	while (left > 0) {
		count++;
		if (pbkdf2_sha1_f(passphrase, length, ssid, ssid_len, 4096, count, digest))
			return -1;
		plen = left > SHA1_MAC_LEN ? SHA1_MAC_LEN : left;
		result += os_memcmp(digest, pos, plen);
		pos += plen;
		left -= plen;
	}

	return result;
}


__kernel void startCrack(__constant char* hash, __constant uchar* ssid, size_t ssid_len, __constant char* passphrase, size_t key_len, __global int* output) {
    char result;
    if(output[3] == 255) return;
    result = pbkdf2_sha1(passphrase, key_len, ssid, ssid_len, hash);
    if(result == 2) {
        output[3] = 255;
        output[0] = get_global_id(0);
        output[1] = get_global_id(1);
        output[2] = get_global_id(2);
    }
    else if(result == -1) output[3] = 127;
    return;
}
