#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <CL/cl.h>
#define REFLECTOR_SIZE 64
#define RING_BUFFER_SIZE 15 // highest index of the ring buffer
#define PROGRAM_NAME "crackalg.cl"
#define SOURCE_LENGTH 2048

typedef struct {
	char* pwstring; //the static password string for this job.
} job;


typedef struct {
	cl_device_id devid; //device id identifying the device in OpenCL
	unsigned volatile char current; // the first job in the ring buffer
	unsigned volatile char empty; // the first empty spot in the ring buffer
	cl_command_queue command_queue;
	job* ringbuffer[RING_BUFFER_SIZE + 1]; // queue for calculated jobs
	cl_mem input;
    cl_mem output;
    cl_mem hash;
    cl_mem ssid;
    cl_kernel kernel;
} device;

const char* reflector = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcedfghijklmnopqrstuvwxyz?@";
const unsigned char reverse_reflector[128] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				   0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 0, 0, 0, 0, 62,
				   63, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
				   25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 0, 0, 0, 0, 0,
				   0, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50,
				   51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 0, 0, 0, 0, 0};
int pwsize;
char* volatile oldstring;
char* hash;
char* ssid;
int ssid_len;
int nr_of_devices;
device* devdb;
pthread_mutex_t strlock = PTHREAD_MUTEX_INITIALIZER;
cl_context contexts[8];
cl_uint nr_of_platforms;
char** programsource;
size_t lineLength[SOURCE_LENGTH];
cl_uint programLength;

void generate_hex_string(const char* ascii_string) {
    int i;
    unsigned char buffer = 0;
    int j = strlen(ascii_string);
    hash = malloc(j);
    for(i = 0; i < j; i++) {
        switch(ascii_string[i]) {
            case '0':
                buffer = 0x0 | buffer;
                break;
            case '1':
                buffer = 0x1 | buffer;
                break;
            case '2':
                buffer = 0x2 | buffer;
                break;
            case '3':
                buffer = 0x3 | buffer;
                break;
            case '4':
                buffer = 0x4 | buffer;
                break;
            case '5':
                buffer = 0x5 | buffer;
                break;
            case '6':
                buffer = 0x6 | buffer;
                break;
            case '7':
                buffer = 0x7 | buffer;
                break;
            case '8':
                buffer = 0x8 | buffer;
                break;
            case '9':
                buffer = 0x9 | buffer;
                break;
            case 'a':
            case 'A':
                buffer = 0xa | buffer;
                break;
            case 'b':
            case 'B':
                buffer = 0xb | buffer;
                break;
            case 'c':
            case 'C':
                buffer = 0xc | buffer;
                break;
            case 'd':
            case 'D':
                buffer = 0xd | buffer;
                break;
            case 'e':
            case 'E':
                buffer = 0xe | buffer;
                break;
            case 'f':
            case 'F':
                buffer = 0xf | buffer;
                break;
            default:
                printf("unsupported character in hex string! check your strings! \r\n");
                exit(1);
        }
        if(i & 1) {
            hash[i>>1] = buffer;
            buffer = 0;
        } else {
            buffer = buffer << 4;
        }
    }
    hash[i>>1] = 0;
}

void get_characters(const char* id, unsigned char* characters) {
    characters[0] = reflector[id[0] & 63];
    characters[1] = reflector[id[1] & 63];
    characters[2] = reflector[id[2] & 63];
    characters[3] = reflector[(id[0] >> 6) | ((id[1] >> 4) & 0xC) | ((id[2] >> 2) & 0x30)];
}

void get_ids(char* ids, const char* characters) {
    ids[0] = characters[0] | ((characters[3] << 6) & 0xC0);
    ids[1] = characters[1] | ((characters[3] << 4) & 0xC0);
    ids[2] = characters[2] | ((characters[3] << 2) & 0xC0);
}

void readFile(FILE *file) {
    int i = 0;
    int j;
    char* test;
    if (file == NULL) {
        printf("Error: file pointer is null.");
        exit(1);
    }
    do {
        test = fgets(programsource[i++], 512, file);
    } while(test != NULL);
    for(j = 0; j < i; j++) lineLength[j] = strlen(programsource[j]);
    programLength = i;
}

bool add_to_device(device* dev, job* j)  {
	if(((dev->empty + 1) & RING_BUFFER_SIZE) == dev->current) return false;
	dev->ringbuffer[dev->empty] = j;
	dev->empty = (dev->empty + 1) & RING_BUFFER_SIZE;
    return true;
}

job* get_next_job(device* dev) {
    job* j;
	if(dev->empty == dev->current) return NULL;
    j = dev->ringbuffer[dev->current];
    dev->current = (dev->current + 1) & RING_BUFFER_SIZE;
    return j;
}

/**
 * this function generates the next password string, if possible
 *
 **/
bool prepare_next_string(device* dev) {
	int i;
    char* newstring;
    if(pwsize == 0) return false;
    newstring = (char*) malloc(pwsize + 1);
    job* newjob = (job*) malloc(sizeof(job));
    pthread_mutex_lock(&strlock);
	memcpy(newstring, oldstring, pwsize);
    newjob->pwstring = newstring;
    for(i = pwsize - 1; i > 3; i--) {
        if(newstring[i] != reflector[REFLECTOR_SIZE - 1]) break;
    }
    if(i == 3) {
        free(newstring);
        free(newjob);
        return false;
    }
    newstring[i] = reflector[reverse_reflector[newstring[i]] + 1];
    i++;
    while(i < pwsize) newstring[i++] = reflector[0];
   	memcpy(oldstring, newstring, pwsize);
    pthread_mutex_unlock(&strlock);
    return add_to_device(dev, newjob);
}

bool cleanup_job(job* job_finished) {
    free(job_finished->pwstring);
    return true;
}

void initialize_device(cl_device_id device, cl_context contextid, cl_program program) {
    char completion[128];
    devdb[nr_of_devices].devid = device;
    devdb[nr_of_devices].current = 0;
    devdb[nr_of_devices].empty = 0;
    devdb[nr_of_devices].kernel =  clCreateKernel ( program, "startCrack", NULL);
    devdb[nr_of_devices].command_queue = clCreateCommandQueue(contextid, device, 0, NULL);
    devdb[nr_of_devices].input = clCreateBuffer( contextid, CL_MEM_READ_ONLY, pwsize, NULL, NULL);
  	devdb[nr_of_devices].output = clCreateBuffer( contextid, CL_MEM_READ_WRITE, 16, NULL, NULL);
    devdb[nr_of_devices].hash = clCreateBuffer(contextid, CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, strlen(hash), hash, NULL);
    devdb[nr_of_devices].ssid = clCreateBuffer(contextid, CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, strlen(ssid), ssid, NULL);
    clGetDeviceInfo(device, CL_DEVICE_NAME, 128, completion, NULL);
    printf("attention! cl device %s is working in the mines from now ... RIP \r\n", completion);

    clSetKernelArg ( devdb[nr_of_devices].kernel, 0, sizeof(cl_mem), &devdb[nr_of_devices].hash);
    clSetKernelArg ( devdb[nr_of_devices].kernel, 1, sizeof(cl_mem), &devdb[nr_of_devices].ssid);
    clSetKernelArg ( devdb[nr_of_devices].kernel, 2, 1, &ssid_len);
    clSetKernelArg ( devdb[nr_of_devices].kernel, 3, sizeof(cl_mem), &devdb[nr_of_devices].input);
    clSetKernelArg ( devdb[nr_of_devices].kernel, 4, 1, &pwsize);
    clSetKernelArg ( devdb[nr_of_devices].kernel, 5, sizeof(cl_mem), &devdb[nr_of_devices].output);
    nr_of_devices++;
}

void cleanup_device(device* dev) {
    clReleaseMemObject(dev->input);
    clReleaseMemObject(dev->output);
    clReleaseMemObject(dev->hash);
    clReleaseMemObject(dev->ssid);
    clReleaseKernel(dev->kernel);
    clReleaseCommandQueue(dev->command_queue);
}

void initialize_platforms() {
    cl_uint i,j;
    cl_int ret;
    cl_uint nr_of_cl_devices[8];
    cl_platform_id platforms[8];
    char logger[65536];
    cl_program program;
    cl_device_id devices[8][8];
    FILE *fp;
    cl_context_properties part_props[3] = { CL_CONTEXT_PLATFORM, 0, 0 };
    ssid_len = strlen(ssid);
    programsource = (char**) malloc(SOURCE_LENGTH*sizeof(char*));
    for(i = 0; i < SOURCE_LENGTH; i++) programsource[i] = malloc(512*sizeof(char));
    fp = fopen(PROGRAM_NAME, "rb");
    readFile(fp);
    fclose(fp);
    clGetPlatformIDs(8, platforms, &nr_of_platforms);
    for(i = 0; i < nr_of_platforms; i++) {
        clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_ALL, 8, devices[i], &nr_of_cl_devices[i]);
        part_props[1] = (cl_context_properties) platforms[i];
        contexts[i] = clCreateContext(part_props, nr_of_cl_devices[i], devices[i], NULL, NULL, &ret);
        program = clCreateProgramWithSource (contexts[i], programLength, (const char **) programsource, lineLength, &ret);
        ret = clBuildProgram(program, nr_of_cl_devices[i], devices[i], "-w -cl-opt-disable", NULL, NULL);
        ret = clGetProgramBuildInfo (program, devices[i][0], CL_PROGRAM_BUILD_LOG, 65536, logger, NULL);
        printf("%s", logger);
        for(j = 0; j < nr_of_cl_devices[i]; j++) {
            initialize_device(devices[i][j], contexts[i], program);
        }
    }
    for(i = 0; i < SOURCE_LENGTH; i++) free(programsource[i]);
    free(programsource);
}

void cleanup_platforms() {
    cl_uint i;
    for(i = 0; i < nr_of_devices; i++) cleanup_device(&devdb[i]);
    for(i = 0; i < nr_of_platforms; i++) clReleaseContext(contexts[i]);
}
    
void analyze(const int ret, const char* source) {
    if(ret == CL_SUCCESS) printf(" %s success \r\n", source);
    if(ret == CL_INVALID_COMMAND_QUEUE) printf("if command_queue is not a valid command-queue");
    if(ret == CL_INVALID_CONTEXT) printf("if the context associated with command_queue and buffer are not the same or if the context associated with command_queue and events in event_wait_list are not the same.");
    if(ret == CL_INVALID_MEM_OBJECT) printf("if buffer is not a valid buffer object.");
    if(ret == CL_INVALID_VALUE) printf(" if the region being read specified by (offset, cb) is out of bounds or if ptr is a NULL value.");
    if(ret == CL_INVALID_EVENT_WAIT_LIST) printf("if event_wait_list is NULL and num_events_in_wait_list greater than 0, or event_wait_list is not NULL and num_events_in_wait_list is 0, or if event objects in event_wait_list are not valid events.");
    if(ret == CL_MEM_OBJECT_ALLOCATION_FAILURE) printf("if there is a failure to allocate memory for data store associated with buffer.");
}
void* worker(void* device_obj) {
    const size_t workborder[3] = {256,256,256};
    device* dev = (device*) device_obj;
    job* job_processed;
    int ret;
    cl_event task_events[2];
    unsigned int answer[4];
    char temp_buffer[3];
    while(pwsize != 0) {
        while(dev->current == dev->empty) sleep(1);
            if(dev->current != ((dev->empty + 1) & RING_BUFFER_SIZE)) if(!prepare_next_string(dev)) { //add an additional object to the ring buffer
                pwsize = 0;
                continue;
            }
            job_processed = get_next_job(dev);
            clEnqueueWriteBuffer(dev->command_queue, dev->input, CL_FALSE, 0, pwsize, job_processed->pwstring, 0, NULL, &task_events[0]);
            ret = clEnqueueNDRangeKernel (dev->command_queue, dev->kernel, 3, NULL, workborder, NULL, 1, task_events, &task_events[1]);
            analyze(ret, "Kernel");
            ret = clEnqueueReadBuffer (dev->command_queue, dev->output, CL_TRUE, 0, 16, answer, 1, &(task_events[1]), NULL);
            analyze(ret, "ReadBuffer");
            if(answer[3] == 255) {
                //yes, we found it!
                temp_buffer[0] = (char) answer[0];
                temp_buffer[1] = (char) answer[1];
                temp_buffer[2] = (char) answer[2];
                get_characters(temp_buffer, job_processed->pwstring);
                job_processed->pwstring[pwsize] = 0;
                printf("result: %s \r\n", job_processed->pwstring);
                pwsize = 0;
            }
            cleanup_job(job_processed);
    }
    exit(0);
    return NULL;
}



int main(int argc, char** argv) {
    pthread_t* worker_thread;
    int i;
    job initial;
    pwsize = atoi(argv[1]);
    generate_hex_string(argv[2]);
    ssid = argv[3];
    printf("password size: %d \r\n", pwsize);
    printf("hash string: %s \r\n", hash);
    printf("SSID string: %s \r\n", ssid);
    oldstring = malloc(pwsize + 1);
    oldstring[pwsize] = '\0';
    for(i = 0; i < pwsize; i++) oldstring[i] = reflector[0];
    nr_of_devices = 0;
    devdb = malloc(64 * sizeof(device));
    nr_of_platforms = 0;
    initialize_platforms();
    worker_thread = malloc(nr_of_devices * sizeof(pthread_t));
//    for(i = 0; i < 1; i++) {
    for(i = 0; i < nr_of_devices; i++) {
        /* create a worker thread which executes the jobs on the device(&x) */
        if(pthread_create(&worker_thread[i], NULL, worker, &devdb[i])) {
            fprintf(stderr, "Error creating thread %d \r\n", i);
            return 1;
        }
    }
    initial.pwstring = malloc(pwsize);
    memcpy(initial.pwstring, oldstring, pwsize);
    add_to_device(devdb, &initial);
    for(i = 0; i < nr_of_devices; i++) {
        prepare_next_string(&devdb[i]);
    }
    for(i = 0; i < nr_of_devices; i++) {
        if(pthread_join(worker_thread[i], NULL)) {
            fprintf(stderr, "Error joining thread\n");
            return 2;
        }
    }
    cleanup_platforms();
    return 0;
}
